% "Copyright (c) 2008 Stanford University. All rights reserved.
% 
% Permission to use, copy, modify, and distribute this software and
% its documentation for any purpose, without fee, and without written
% agreement is hereby granted, provided that the above copyright
% notice, the following two paragraphs and the author appear in all
% copies of this software.
% 
% IN NO EVENT SHALL STANFORD UNIVERSITY BE LIABLE TO ANY PARTY FOR
% DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
% ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
% IF STANFORD UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
% DAMAGE.
% 
% STANFORD UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES,
% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
% PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND STANFORD UNIVERSITY
% HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
% ENHANCEMENTS, OR MODIFICATIONS."
% 
%  
% Main test script
% 
% @author HyungJune Lee, Martin Wicke, Branislav Kusy, Leonidas Guibas 
%         (hyungjune.lee@gmail.com)
% 
% @date   Nov 7 2008
% 
% parameters:
%   train_burst_size: the number of samples per burst used in training
%   test_burst_size: the number of samples per burst used in testing
%   data_type: 2=RSSI, 3=PRR, 4=LQI
%   hist_size: length of history including the current one, 
%              i.e., hist_size = k + 1 where k is the length of the previous history 
%              as in the definition of the MELT'08 paper
%   average_flag: average samples before fitting
%   korobov_flag: 0=use subsampling of paths for centers; 1=use korobov sampling
%   center_numbers:  subsample every x'th sample (if koborov==0)
%   center_numbers:  number of centers to use (if koborov==1)
% 
%   test_set: test set vectors consisting of 1=exp1, 7=exp7, 8=exp8, 9=exp9

close all;
clear all;

global data_path;
global result_path;
global intermediate_path;
global do_single;

data_path = 'data/';
result_path = 'results/';
intermediate_path = 'intermediate/';

train_burst_sizes = [5];
test_burst_sizes = [5];
data_types = [2]; % allowed 2 3 4
hist_sizes = [5]; 
average_flags = [0]; % allowed 0 1
koborov_flags = [0]; % allowed 0 1
center_numbers = [0.1]; % ratio if koborov == 0, number otherwise

test_sets = [1 7 8 9]; % allowed 1 7 8 9

% single precision computations for big problems
do_single = false;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nodes = 11;

if do_single
    disp(['Using single precision for RBF matrix solve.']);
end

for data_type = data_types    

    switch (data_type)
        case 2
            data_type_name = 'rssi';
        case 3
            data_type_name = 'prr';
        case 4
            data_type_name = 'lqi';
    end

    for hist_size = hist_sizes
        for train_burst_size = train_burst_sizes
            for average_flag = average_flags
                for koborov_flag = koborov_flags
                    for numcenters = center_numbers

                        % at this point, all parameters for the model are fixed.

                        rbfname = [intermediate_path ...
                              'rbfcoeff-' data_type_name...
                              '-hs' int2str(hist_size)...
                              '-b' int2str(train_burst_size)...
                              '-avg' int2str(average_flag)...
                              '-k' int2str(koborov_flag)...
                              '-s' num2str(numcenters)...
                              '.mat'];

                        if ~exist(rbfname, 'file')
                            disp(['cannot open "' rbfname '", creating model first']);
                            [rbfcoeff, centers] = prepare(data_type, hist_size, train_burst_size, average_flag, koborov_flag, numcenters);
                            [dim, m] = size(centers);
                            save(rbfname, 'rbfcoeff', 'centers');    
                        else
                            load(rbfname, 'rbfcoeff', 'centers');

                            n = length(rbfcoeff(1).rbfcoeff);
                            [dim, m] = size(centers);
                            
                            disp(['Using ' num2str(m) ' centers.']);

                            if n ~= m+dim+1
                                error('different numbers of centers and coefficients.');
                            end

                            if dim ~= nodes * hist_size
                                error(['Dimension (' num2str(dim) ') is not what we expect(' num2str(nodes*hist_size) ').']);
                            end
                        end
                        
                        for test_burst_size = test_burst_sizes

                        % loop over test sets is in eval_error
                        [pos_error, path_error] = eval_error(data_type, hist_size, train_burst_size, test_burst_size, average_flag, koborov_flag, numcenters, rbfcoeff, centers, test_sets);

                        disp(['Data=' data_type_name...
                              ', ' num2str(m) ' centers'...
                              ', hist_size=' int2str(hist_size)...
                              ', train burst=' int2str(train_burst_size)...
                              ', avg=' int2str(average_flag)...
                              ', koborov=' int2str(koborov_flag)...
                              ', centers=' num2str(numcenters)...
                              ', test burst=' int2str(test_burst_size)]);
                        disp(['pos error:  mean=' num2str(mean(pos_error)) ', quantiles: ']);
                        disp(quantile(pos_error,[.025 .25 .50 .75 .975]));
                        disp(['path error: mean=' num2str(mean(path_error)) ', quantiles: ']);
                        disp(quantile(path_error,[.025 .25 .50 .75 .975]));
                        disp(['using ' int2str(length(pos_error)) ' samples.']);

                        end
                    end
                end
            end
        end
    end
end
