% "Copyright (c) 2008 Stanford University. All rights reserved.
% 
% Permission to use, copy, modify, and distribute this software and
% its documentation for any purpose, without fee, and without written
% agreement is hereby granted, provided that the above copyright
% notice, the following two paragraphs and the author appear in all
% copies of this software.
% 
% IN NO EVENT SHALL STANFORD UNIVERSITY BE LIABLE TO ANY PARTY FOR
% DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
% ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
% IF STANFORD UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
% DAMAGE.
% 
% STANFORD UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES,
% INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
% MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
% PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND STANFORD UNIVERSITY
% HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
% ENHANCEMENTS, OR MODIFICATIONS."
% 
%  
% This function prepares an RBF fit for a specific experimental setting
% 
% @author HyungJune Lee, Martin Wicke, Branislav Kusy, Leonidas Guibas
%         (hyungjune.lee@gmail.com)
% 
% @date   Nov 7 2008
% 
% usage:
%
% use koborov:
% [rbfcoeff, centers] = prepare(data_type, prev_k, burst_size, average, 1, numcenters_or_subsampling)
% don't use koborov
% [rbfcoeff, centers] = prepare(data_type, prev_k, burst_size, average, 0, numcenters_or_subsampling)
%
% returns:
%   rbfcoeff: RBF fitting coefficients
%   centers: RBF centers for evaluation
%
% parameters:
%   data_type: 2=RSSI, 3=PRR, 4=LQI
%   prev_k: length of history
%   burst_size: number of samples per burst used
%   average: average samples before fitting
%   koborov_flag: 0=use subsampling of paths for centers; 1=use koborov sampling
%   numcenters_or_subsampling:  ratio of samples used as centers (if koborov_flag==0)
%   numcenters_or_subsampling:  number of centers to use (if koborov_flag==1)

function [rbfcoeff, centers] = prepare(data_type, prev_k, burst_size, average, koborov_flag, numcenters_or_subsampling)

global intermediate_path;

nodes = 11;

switch (data_type)
    case 2
        data_type_name = 'rssi';
    case 3
        data_type_name = 'prr';
    case 4
        data_type_name = 'lqi';
end

dataname = [intermediate_path 'constraints-' data_type_name...
            '-hs' int2str(prev_k)...
            '-b' int2str(burst_size)...
            '-avg' int2str(average)...
            '.mat'];


if ~exist(dataname, 'file')
    disp(['cannot open "' dataname '", generating constraints...']);
    [X_rbf, Y_rbf] = assemble_constraints(data_type, prev_k, burst_size, average);
    save(dataname, 'X_rbf', 'Y_rbf');    
else
    disp(['loading constraints from file.']);
    load(dataname, 'X_rbf', 'Y_rbf');
    [dim, n] = size(X_rbf);
    [dim2, n2] = size(Y_rbf);

    if n ~= n2
        error('number of constraints inconsistent.');
    end

    if dim ~= nodes * prev_k
        error(['Dimension (' num2str(dim) ') is not what we expect(' num2str(nodes*prev_k) ').']);
    end
end

[dim, n] = size(X_rbf);
disp(['Using ' int2str(n) ' constraints of dimension ' int2str(dim) '.']);

% do the actual fit

% select centers 
centername = [intermediate_path 'centers-' data_type_name...
              '-hs' int2str(prev_k)...
              '-b' int2str(burst_size)...
              '-avg' int2str(average)...
              '-k' int2str(koborov_flag)...
              '-s' num2str(numcenters_or_subsampling)...
              '.mat'];

if ~exist(centername, 'file')
    disp(['cannot open "' centername '", selecting centers...']);
    if koborov_flag
        centers = MIN_RSSI * ones(nodes*prev_k, numcenters_or_subsampling) + koborov(nodes*prev_k, numcenters_or_subsampling) * (MAX_RSSI - MIN_RSSI);
    else
        centers = select_centers(numcenters_or_subsampling, X_rbf);
    end
    save(centername, 'centers');    
else
    disp('loading centers from file.');
    load(centername, 'centers');
    [dim, n] = size(centers);

    if dim ~= nodes * prev_k
        error(['Center dimension (' num2str(dim) ') is not what we expect(' num2str(nodes*prev_k) ').']);
    end
end

[dim,n] = size(X_rbf);
[dim,m] = size(centers);
disp(['Computing RBF approximation: ' int2str(n) ' constraints, ' int2str(m) ' centers.']);

for i = 1:2
    if (i == 1)
        char = 'x';
    else
        char = 'y';
    end
    
    disp(['solving ' char '...']);
    
    savename = [intermediate_path 'rbfcoeff-' char '-' data_type_name...
            '-hs' int2str(prev_k)...
            '-b' int2str(burst_size)...
            '-avg' int2str(average)...
            '-k' int2str(koborov_flag)...
            '-s' num2str(numcenters_or_subsampling)];
    
    rbfcoeff(i) = rbfcreate_k(X_rbf, Y_rbf(i,:), centers, ...
                              'RBFFunction', 'multiquadric', ...
                              'RBFConstant', 1, ...
                              'RBFSmooth', 0.1, ...
                              'Savename', savename);
end

disp('  done computing RBF approximation.');

end





function [X_rbf Y_rbf] = assemble_constraints(data_type, prev_k, burst_size, average)

global data_path;

geo = [];
melt_map;

dt = 0.6666;
MAX_RSSI = -45;
MIN_RSSI = -95;
MIN_LQI = 50;
nodes = 11;

test_file = ['m_exp2_01.txt'; 'm_exp2_02.txt'; 'm_exp2_03.txt'; 'm_exp2_04.txt'; 'm_exp2_05.txt';
             'm_exp3_01.txt'; 'm_exp3_02.txt'; 'm_exp3_03.txt'; 'm_exp3_04.txt';
             'm_exp4_01.txt'; 'm_exp4_02.txt'; 'm_exp4_03.txt'; 'm_exp4_04.txt'; 'm_exp4_05.txt';
             'm_exp5_01.txt'; 'm_exp5_02.txt'; 'm_exp5_03.txt'; 'm_exp5_04.txt'; 'm_exp5_05.txt';
             'm_exp6_01.txt'; 'm_exp6_02.txt'; 'm_exp6_03.txt'; 'm_exp6_04.txt'; 'm_exp6_05.txt'];
time_file = ['c_exp2_01.txt'; 'c_exp2_02.txt'; 'c_exp2_03.txt'; 'c_exp2_04.txt'; 'c_exp2_05.txt';
             'c_exp3_01.txt'; 'c_exp3_02.txt'; 'c_exp3_03.txt'; 'c_exp3_04.txt';
             'c_exp4_01.txt'; 'c_exp4_02.txt'; 'c_exp4_03.txt'; 'c_exp4_04.txt'; 'c_exp4_05.txt';
             'c_exp5_01.txt'; 'c_exp5_02.txt'; 'c_exp5_03.txt'; 'c_exp5_04.txt'; 'c_exp5_05.txt';
             'c_exp6_01.txt'; 'c_exp6_02.txt'; 'c_exp6_03.txt'; 'c_exp6_04.txt'; 'c_exp6_05.txt'];
landmark_file = ['s_exp2.txt'; 's_exp2.txt'; 's_exp2.txt'; 's_exp2.txt'; 's_exp2.txt';
                 's_exp3.txt'; 's_exp3.txt'; 's_exp3.txt'; 's_exp3.txt';
                 's_exp4.txt'; 's_exp4.txt'; 's_exp4.txt'; 's_exp4.txt'; 's_exp4.txt';
                 's_exp5.txt'; 's_exp5.txt'; 's_exp5.txt'; 's_exp5.txt'; 's_exp5.txt';
                 's_exp6.txt'; 's_exp6.txt'; 's_exp6.txt'; 's_exp6.txt'; 's_exp6.txt'];

% test_file = ['m_exp2_01.txt'; 'm_exp2_02.txt'];
% time_file = ['c_exp2_01.txt'; 'c_exp2_02.txt'];
% landmark_file = ['s_exp2.txt'; 's_exp2.txt'];

test_file = horzcat(repmat(data_path, length(test_file(:,1)), 1), test_file);
time_file = horzcat(repmat(data_path, length(time_file(:,1)), 1), time_file);
landmark_file = horzcat(repmat(data_path, length(landmark_file(:,1)), 1), landmark_file);

X_rbf = [];
Y_rbf = [];

ds_size = length(test_file(:,1));

for ds = 1:ds_size
    
    clear time;
    disp(['Reading data ' test_file(ds,:)]);

    fdata = load(test_file(ds,:));
    fdata = fdata(:,[1 2 4 5]);
    [m n] = size(fdata);

    disp(['  reading times ' time_file(ds,:)]);

    time = load(time_file(ds,:));
    time = time - time(1);
    time = round(time/1000/dt)+1;
    time = time(2:end);

    disp(['  reading landmarks ' landmark_file(ds,:)]);

    landmark = load(landmark_file(ds,:));

    disp(['  preprocessing...']);

    for i=1:m
        if (i==1)
            tmp_index = fdata(i,1);
            index = [tmp_index i];
        elseif (fdata(i,1) ~= tmp_index)
                tmp_index = fdata(i,1);
                index = [index ; tmp_index i];
        end
        if (fdata(i,3) >= 128)
            fdata(i,3) = fdata(i,3) - 256 - 45;
        else
            fdata(i,3) = fdata(i,3) - 45;
        end
    end
    max_x = max(fdata(:,2));
    
    max_sec = floor(max_x/30)+1;
    
    if (time(end) > max_sec)
        time(end) = max_sec;
    end
    
    tmp_count = 0;
    len = nodes*max_sec;
    rdata = [];
    disp_max_x = ceil(dt*max_sec) + 3;

    for i=1:nodes
        rdata = [rdata; i*ones(max_sec,1) zeros(max_sec, 3)];
    end

    for i=1:m
        if (mod(fdata(i,2),30) >= burst_size)
            continue;
        end
        pt = (fdata(i,1)-1)*max_sec + (floor(fdata(i,2)/30)+1);
        if average
            sub_seq = 1;
        else
            sub_seq = mod(fdata(i,2),30)+1;
        end
        rdata(pt,1,sub_seq) = fdata(i,1);
        rdata(pt,2,sub_seq) = rdata(pt,2,sub_seq) + fdata(i,3);
        rdata(pt,3,sub_seq) = rdata(pt,3,sub_seq) + 1;
        rdata(pt,4,sub_seq) = rdata(pt,4,sub_seq) + fdata(i,4);
    end
    clear fdata; % by hjlee
    
    tmp_prr = zeros(len,1);
    
    for i=1:len
        if average
            j=1;
            if (rdata(i,3,j) == 0)
                rdata(i,2,j) = MIN_RSSI;
                rdata(i,4,j) = MIN_LQI;
            else
                rdata(i,2,j) = rdata(i,2,j)/rdata(i,3,j);
                rdata(i,4,j) = rdata(i,4,j)/rdata(i,3,j);
                rdata(i,3,j) = rdata(i,3,j)/burst_size;
            end
        else
            for j=1:burst_size
                if (rdata(i,3,j) == 0)
                    rdata(i,2,j) = MIN_RSSI;
                    rdata(i,4,j) = MIN_LQI;
                end
            end
        end
    end
    
    clear data;
    for i=1:nodes
        if average
            data(i,:,1) = [rdata((i-1)*max_sec+1:i*max_sec,data_type,1)'];
        else
            for j=1:burst_size
                data(i,:,j) = [rdata((i-1)*max_sec+1:i*max_sec,data_type,j)'];
            end
        end
    end
    clear rdata; % by hjlee
    
    disp(['  creating constraints for ' int2str((time(end)-time(1))) ' bursts.']);
    %FIXME: this should be preallocated
%     X_rbf = []; % This should be moved to the beginning by HyungJune
%     Y_rbf = [];
    
    for i=time(1):time(end)
        if (i < prev_k)
            continue;
        end
        
        for k = 1:burst_size
            tmpX = [];
            tmpY = zeros(2,1);

            if average
                for j=1:nodes
                    tmpX = [tmpX; data(j, i-prev_k+1:i,1)'];
                end
            else
                for j=1:nodes
                    tmpX = [tmpX; data(j, i-prev_k+1:i,k)'];
                end
            end
            X_rbf = [X_rbf tmpX];

            time_sc_e = find(time == i);
            if (~isempty(time_sc_e))
                tmpY = geo(landmark(time_sc_e(1)),2:3)';
            else
                time_sc_ie = find(time > i);
                next_loc = geo(landmark(time_sc_ie(1)),2:3)';
                prev_loc = geo(landmark(time_sc_ie(1)-1),2:3)';
                % FIXME: fix time interpolation
                tmpY = (i-time(time_sc_ie(1)-1))/(time(time_sc_ie(1))-time(time_sc_ie(1)-1))*(next_loc-prev_loc)+prev_loc;
            end
            Y_rbf = [Y_rbf tmpY];

            if average
                break;
            end
        end
    end  
end

clear data;

disp('  done reading training data.');

end
