function options = rbfcreate_k(x, y, c, varargin)

global do_single;

%RBFCREATE Creates an RBF interpolation
%   OPTIONS = RBFSET(X, Y, C, 'NAME1',VALUE1,'NAME2',VALUE2,...) creates an   
%   radial base function interpolation 
%   
%   RBFCREATE with no input arguments displays all property names and their
%   possible values.
%   
%   RBFCREATE PROPERTIES
% 
% Alex Chirokov, alex.chirokov@gmail.com
% 16 Feb 2006
% 
% Modified by HyungJune Lee, Martin Wicke, Branislav Kusy, Leonidas Guibas 
%             (hyungjune.lee@gmail.com)

tic;

% Print out possible values of properties.
if (nargin == 0) & (nargout == 0)
  fprintf('               x: [ dim by n matrix of coordinates for the constraints ]\n');
  fprintf('               y: [   1 by n vector of values at nodes ]\n');
  fprintf('               c: [ dim by m matrix of coordinates for the centers ]\n');
  fprintf('     RBFFunction: [ gaussian  | thinplate | cubic | multiquadrics | {linear} ]\n');
  fprintf('     RBFConstant: [ positive scalar     ]\n');
  fprintf('       RBFSmooth: [ positive scalar {0} ]\n');
  fprintf('           Stats: [ on | {off} ]\n');
  fprintf('        Savename: [ filename suffix for save files ]\n');
  fprintf('\n');
  return;
end
Names = [
    'RBFFunction      '
    'RBFConstant      '
    'RBFSmooth        '
    'Stats            '
    'Savename         '
];
[m,n] = size(Names);
names = lower(Names);

options = [];
for j = 1:m
  options.(deblank(Names(j,:))) = [];
end

%**************************************************************************
%Check input arrays 
%**************************************************************************
[nXDim nXCount]=size(x);
[nYDim nYCount]=size(y);

if (nXCount~=nYCount)
  error(sprintf('x and y should have the same number of rows'));
end;

if (nYDim~=1)
  error(sprintf('y should be n by 1 vector'));
end;

%options.('x')           = x;
%options.('y')           = y;
%options.('c')           = c;
%**************************************************************************
%Default values 
%**************************************************************************
options.('RBFFunction') = 'linear';
options.('RBFConstant') = (prod(max(x')-min(x'))/nXCount)^(1/nXDim); %approx. average distance between the nodes 
options.('RBFSmooth')   = 0;
options.('Stats')       = 'off';
options.('Savename')    = '';

%**************************************************************************
% Argument parsing code: similar to ODESET.m
%**************************************************************************

i = 1;
% A finite state machine to parse name-value pairs.
if rem(nargin-3,2) ~= 0
  error('Arguments must occur in name-value pairs.');
end
expectval = 0;                          % start expecting a name, not a value
while i <= nargin-3
  arg = varargin{i};
    
  if ~expectval
    if ~isstr(arg)
      error(sprintf('Expected argument %d to be a string property name.', i));
    end
    
    lowArg = lower(arg);
    j = strmatch(lowArg,names);
    if isempty(j)                       % if no matches
      error(sprintf('Unrecognized property name ''%s''.', arg));
    elseif length(j) > 1                % if more than one match
      % Check for any exact matches (in case any names are subsets of others)
      k = strmatch(lowArg,names,'exact');
      if length(k) == 1
        j = k;
      else
        msg = sprintf('Ambiguous property name ''%s'' ', arg);
        msg = [msg '(' deblank(Names(j(1),:))];
        for k = j(2:length(j))'
          msg = [msg ', ' deblank(Names(k,:))];
        end
        msg = sprintf('%s).', msg);
        error(msg);
      end
    end
    expectval = 1;                      % we expect a value next
    
  else
    options.(deblank(Names(j,:))) = arg;
    expectval = 0;      
  end
  i = i + 1;
end

if expectval
  error(sprintf('Expected value for property ''%s''.', arg));
end

    
%**************************************************************************
% Creating RBF Interpolation
%**************************************************************************

switch lower(options.('RBFFunction'))
      case 'linear'          
        options.('rbfphi')   = @rbfphi_linear;
      case 'cubic'
        options.('rbfphi')   = @rbfphi_cubic;
      case 'multiquadric'
        options.('rbfphi')   = @rbfphi_multiquadrics;
      case 'thinplate'
        options.('rbfphi')   = @rbfphi_thinplate;
      case 'gaussian'
        options.('rbfphi')   = @rbfphi_gaussian;
    otherwise
        error('Unknown function type');
end

phi = options.('rbfphi');

save_requested = ~strcmp(options.('Savename'), '');

if save_requested
    disp(['Using save files when possible, prefix: ' options.('Savename')]);
end

if save_requested && exist([options.('Savename') '-postsolve.mat'], 'file')
    disp('Loading RBF coefficients from file.');
    load([options.('Savename') '-postsolve.mat'], 'rbfcoeff');
else
    disp('No saved coefficients found.');
    
    if save_requested && exist([options.('Savename') '-presolve.mat'], 'file')
        disp('Loading RBF system matrix from file.');
        load([options.('Savename') '-presolve.mat'], 'A', 'b');       
    else
        disp('Assembling RBF system matrix...');
        A=rbfAssemble(x, c, phi, options.('RBFConstant'), options.('RBFSmooth'));
        if (do_single)
            b=single([y'; zeros(nXDim+1, 1)]);
        else
            b=[y'; zeros(nXDim+1, 1)];
        end
        
        if save_requested
            disp('saving RBF system matrix...');
            save([options.('Savename') '-presolve.mat'], 'A', 'b');
        end
    end

    disp('solving...');
    
    rbfcoeff=A\b;

    %SVD
    %[U,S,V] = svd(A);
    %for i=1:1:nXCount+1
    %    if (S(i,i)>0) S(i,i)=1/S(i,i); end;   
    %end;    
    %rbfcoeff = V*S'*U*b;

    % rbfcoeff = pinv(A)*y';

    %disp(['coefficient vector size: ' int2str(length(rbfcoeff))]);

    if save_requested
        disp('saving RBF coefficients...');
        save([options.('Savename') '-postsolve.mat'], 'rbfcoeff');
    end
end

options.('rbfcoeff') = rbfcoeff;

if (strcmp(options.('Stats'),'on'))
    fprintf('%d point RBF interpolation was created in %e sec\n', length(c), toc);  
    fprintf('\n');
end;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [A]=rbfAssemble(x, c, phi, const, smooth)

global do_single;

[dim n]=size(x);
[dim2 m]=size(c);

if dim2 ~= dim
error('dimensionality of rbf centers and constraints must agree');
end

% A=zeros(n,m+dim+1);
% for i=1:n
%    for j=1:m
%        r=norm(x(:,i)-c(:,j));
%        temp=feval(phi,r,const);
%        A(i,j)=temp;
%    end
%    %A(j,j) = A(j,j) - smooth;
% end
% % Polynomial part
% P = -ones(n,1);
% P=[ones(n,1) x'];
% A = [ A P ];

if do_single
    A=single(zeros(n+dim+1, m+dim+1));
else
    A=zeros(n+dim+1, m+dim+1);
end

tic
msgint = 10;
msgtime = msgint;
for i=1:n
    for j=1:m
        r=norm(x(:,i)-c(:,j));
        temp=feval(phi,r,const);
        A(i,j)=temp;
    end
    t = toc;
    
    if (t > msgtime)
        disp([num2str(i/n*100) '% done...']);
        msgtime = msgtime + msgint;
    end
    %A(j,j) = A(j,j) - smooth;
end
% Polynomial part 
A(1:n, m+1) = ones(n,1);
A(1:n, m+2:end) = x';
A(n+1, 1:m) = ones(1,m);
A(n+2:end, 1:m) = c;

%P=[ones(n,1) x'];
%Q=[ones(1,m); c];
%A = [ A(1:n,1:m)      P
%      Q zeros(dim+1,dim+1)];
%clear P Q;
  
%A=zeros(n,n);
%for i=1:n
%    for j=1:i
%        r=norm(x(:,i)-x(:,j));
%        temp=feval(phi,r, const);
%        A(i,j)=temp;
%        A(j,i)=temp;
%    end
%    A(i,i) = A(i,i) - smooth;
%end
% Polynomial part
%P=[ones(n,1) x'];
%A = [ A      P
%      P' zeros(dim+1,dim+1)];
end

%**************************************************************************
% Radial Base Functions
%************************************************************************** 
function u=rbfphi_linear(r, const)
u=r;
end

function u=rbfphi_cubic(r, const)
u=r.*r.*r;
end

function u=rbfphi_gaussian(r, const)
u=exp(-0.5*r.*r/(const*const));
end

function u=rbfphi_multiquadrics(r, const)
u=sqrt(1+r.*r/(const*const));
end

function u=rbfphi_thinplate(r, const)
u=r.*r.*log(r+1);
end
